/* -*- C++ -*-
 *
 * Copyright (C) 2002 Jean-Daniel Fekete
 * 
 * This file is part of MillionVis.
 * 
 * MillionVis is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 * 
 * MillionVis is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MillionVis; see the file COPYING.  If not, write to the
 * Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA. 
 */
#include <infovis/tree/property_tree.hpp>
#include <boost/directory.h>
#include <sys/stat.h>

namespace infovis {

typedef property_tree Tree;
typedef tree_traits<Tree>::node_descriptor node_descriptor;
typedef Tree::prop_id prop_id;

struct dir_property_tree_builder {
  Tree& tree_;
  Tree::prop_id name_;
  Tree::prop_id size_;
  Tree::prop_id type_;
  Tree::prop_id mtime_;
  Tree::prop_id atime_;
  Tree::prop_id ctime_;
    
  unsigned build(node_descriptor parent, const std::string& dirname) {
    using namespace boost::filesystem;
    unsigned ret = 0;

    for (dir_it it(dirname); it != dir_it(); ++it) {
      if ((*it)[0] == '.')
	continue;
      ret++;
      node_descriptor n = add_node(parent, tree_);
      tree_.get_prop_numeric(size_)[n] = boost::filesystem::get<size>(it);
      tree_.get_prop_numeric(type_)[n] = boost::filesystem::get<is_directory>(it) ? 1 : 0;
      struct stat s;
      stat((dirname+(*it)).c_str(), &s);

      tree_.get_prop_numeric(mtime_)[n] = s.st_mtime;
      tree_.get_prop_numeric(ctime_)[n] = s.st_ctime;
      tree_.get_prop_numeric(atime_)[n] = s.st_atime;

      if (boost::filesystem::get<is_directory>(it)) {
	std::string name = *it + "/";
	tree_.get_prop_string(name_)[n] = name;
	ret += build(n, dirname+name);
      }
      else
	tree_.get_prop_string(name_)[n] = *it;
    }
    return ret;
  }
  dir_property_tree_builder(Tree& t,
			    Tree::prop_id n,
			    Tree::prop_id s,
			    Tree::prop_id tp,
			    Tree::prop_id m,
			    Tree::prop_id c,
			    Tree::prop_id a)
    : tree_(t), name_(n), size_(s), type_(tp),
      mtime_(m), ctime_(c), atime_(a) { }
};

unsigned dir_property_tree(const std::string& dirname,
			   property_tree& t)
{
  Tree::prop_id name = t.add_property("name", Tree::type_string);
  Tree::prop_id size = t.add_property("size", Tree::type_numeric);
  Tree::prop_id type = t.add_property("type", Tree::type_numeric);
  Tree::prop_id mtime = t.add_property("mtime", Tree::type_numeric);
  Tree::prop_id atime = t.add_property("atime", Tree::type_numeric);
  Tree::prop_id ctime = t.add_property("ctime", Tree::type_numeric);
    
  if (num_nodes(t) == 0)
    add_node(root(t), t);
  std::string dname(dirname);
  if (*dname.rbegin() != '/')
    dname.append("/");
  struct stat s;
  stat(dirname.c_str(), &s);
  t.get_prop_string(name)[root(t)] = dname;
  t.get_prop_numeric(size)[root(t)] = s.st_size;
  t.get_prop_numeric(type)[root(t)] = ((s.st_mode&S_IFDIR) != 0);
  t.get_prop_numeric(mtime)[root(t)] = s.st_mtime;
  t.get_prop_numeric(ctime)[root(t)] = s.st_ctime;
  t.get_prop_numeric(atime)[root(t)] = s.st_atime;
  dir_property_tree_builder builder(t, name, size, type, mtime, atime, ctime);
  return builder.build(root(t), dname);
}


} // namespace infovis
